// $Id: ThumbnailPanel.java 16027 2012-07-13 11:18:01Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.viewer2.util.LongRunningTask;
import org.faceless.util.SoftInterruptibleThread;
import org.faceless.pdf2.*;
import javax.swing.*;
import javax.swing.undo.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.font.FontRenderContext;
import java.io.*;
import java.util.*;
import java.beans.*;
import java.util.List;
import java.awt.dnd.*;
import java.awt.datatransfer.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;
import java.awt.event.*;
import java.text.NumberFormat;

/**
 * Creates a {@link SidePanel} that displays the page thumbnails.
 * Since release 2.10.2 the Thumbnail panel can also be used to
 * reorder, delete and insert pages or entire documents. Modifying
 * a document in this way requires the {@link EncryptionHandler} on
 * the document grants the "Assemble" right.
 *
 * <div class="initparams">
 * The following <a href="../doc-files/initparams.html">initialization parameters</a> can be specified to configure this feature.
 * <table>
 * <tr><th>editable</th><td><code>true</code> or <code>false</code>, for {@link #setEditable setEditable()}</td></tr>
 * <tr><th>scrollFollow</th><td><code>true</code> or <code>false</code>, for {@link #setScrollFollow setScrollFollow()}</td></tr>
 * <tr><th>usePageLabels</th><td><code>true</code> or <code>false</code>, for {@link #setUsePageLabels setUsePageLabels()}</td></tr>
 * <tr><th>thumbnailSize</th><td><code>number</code>, for {@link #setThumbnailSize setThumbnailSize()}. Default is 100</td></tr>
 * </table>
 * </div>
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">ShowHideThumbnails</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.8
 */
public class ThumbnailPanel extends SidePanelFactory {

    private boolean editmode, scrollfollow, uselabels = true;
    private int thumbnailsize;

    /**
     * Return a new ThumbnailPanel
     * @deprecated since 2.10.2 just calls the public constructor
     */
    public ThumbnailPanel getInstance() {
        return new ThumbnailPanel();
    }

    /**
     * Create a new ThumbnailPanel
     * @since 2.10.2
     */
    public ThumbnailPanel() {
        super("ShowHideThumbnails");
        setEditable(true);
        setScrollFollow(true);
        setThumbnailSize(100);
    }

    /**
     * Determine whether SidePanels created by this factory allow
     * pages to be reordered, deleted and so on. The default is true.
     * @param editmode true if SidePanels created by this Factory allow alterations to the PDF
     * @since 2.10.2
     */
    public void setEditable(boolean editmode) {
        this.editmode = editmode;
    }

    /**
     * Set whether the thumbnail panel should scroll to follow the
     * currently selected page or not. The default is true.
     * @since 2.11.6
     * @param follow whether to follow the currently selected page
     * in the viewport or not.
     */
    public void setScrollFollow(boolean follow) {
        this.scrollfollow = follow;
    }

    /**
     * Set whether to number the pages using the "page labels" if defined
     * on this PDF, or whether to always use the physical page number.
     * @param uselabels if true, use the page labels if defined, otherwise use the physical page number
     * @see PDF#getPageLabel
     * @see PageNumber#setUsePageLabels
     * @since 2.11.19
     */
    public void setUsePageLabels(boolean uselabels) {
        this.uselabels = uselabels;
    }
    
    /**
     * Set the size of the longest dimension of each thumbnail. The
     * default is 100, although this can be changed by the "thumbnailSize"
     * property or via the user-interface while running the viewer.
     * @since 2.11.13
     */
    public void setThumbnailSize(int thumbnailsize) {
        this.thumbnailsize = thumbnailsize;
    }

    public void initialize(PDFViewer viewer) {
        super.initialize(viewer);
        String val = getFeatureProperty(viewer, "editable");
        if (val!=null) {
            setEditable("true".equals(val));
        }
        val = getFeatureProperty(viewer, "scrollFollow");
        if (val != null) {
            setScrollFollow("true".equals(val));
        }
        val = getFeatureProperty(viewer, "usePageLabels");
        if (val != null) {
            setUsePageLabels("true".equals(val));
        }
        val = getFeatureProperty(viewer, "thumbnailSize");
        int thumbnailsize = 100;
        if (val != null) {
            try {
                thumbnailsize = Math.max(20, Math.min(400, Integer.parseInt(val)));
            } catch (Exception e) {}
        }
        setThumbnailSize(thumbnailsize);
    }

    public SidePanel createSidePanel() {
        return new ThumbnailSidePanel(editmode, scrollfollow, uselabels, thumbnailsize);
    }

    /**
     * The scrollable pane containing the SidePanel. Not much of interest in this class,
     * just a wrapper around ThumbPanel
     */
    static class ThumbnailSidePanel extends JScrollPane implements SidePanel, PDFBackgroundTask {
        private final ThumbPanel view;
        private final boolean factoryeditable;
        private final Icon icon;

        ThumbnailSidePanel(boolean editable, boolean scrollfollow, boolean uselabels, int thumbnailsize) {
            this.factoryeditable = editable;
            view = new ThumbPanel(scrollfollow, uselabels, thumbnailsize, this);
            view.setVisible(false);
            setViewportView(view);
            Util.fixScrollPaneKeyBindings(this);
            icon = new ImageIcon(PDFViewer.class.getResource("resources/icons/page_white_copy.png"));
        }

        public Icon getIcon() {
            return icon;
        }

        public String getName() {
            return "Pages";
        }

        public void setDocumentPanel(DocumentPanel docpanel) {
            view.setDocumentPanel(docpanel, factoryeditable);
        }

        public void panelVisible() {
            view.setVisible(true);
            synchronized (view) {
                view.notifyAll();
            }
            view.requestFocusInWindow();
        }

        public void panelHidden() {
            view.setVisible(false);
        }

        /**
         * Return true if the thumbnails are still rendering
         * @since 2.11.7
         */
        public boolean isRunning() {
            return view.isRunning();
        }

        /**
         * Return true if the Thumbnail rendering has been paused
         * with the {@link #pause} method
         * @since 2.11.7
         */
        public boolean isPaused() {
            return view.isPaused();
        }

        /**
         * Interrupt the thumbnail rendering thread. This method will
         * wait until the thread has completed.
         * @since 2.11.7
         */
        public void pause() throws InterruptedException {
            view.setPaused(true);
        }

        public void unpause() {
            view.setPaused(false);
        }
    }

    /**
     * This class is the JPanel containing all the thumbnails. It serves
     * as a DropTarget for DND, as a the Scrollable content of the ThumbnailSidePanel
     * class above, and handles actions when the document is updated and when a popup
     * has run on the selected panels
     */
    static class ThumbPanel extends JPanel implements Scrollable, DocumentPanelListener, Autoscroll, PropertyChangeListener, DropTargetListener {

        private final ThumbnailSidePanel parent;
        private final boolean scrollfollow, uselabels;
        private final int thumbnailsize;
        private final NumberFormat pageNumberFormat;

        private DocumentPanel docpanel;
        private PDF pdf;
        private volatile Dimension childsize;
        private boolean editable, updatepagepanelsrequired, animatemovements;
        private AnimateThread animatethread;
        private PainterThread painterthread;
        private SinglePagePanel currentPagePanel;       // Last one that was selected - broad meaning
        private SinglePagePanelFactory singlepagepanelfactory;
        private JPopupMenu popup;
        private List pagepanellist;
        private Component dragTarget; // A SinglePagePanel to insert before, or this to insert at end, or null
        private AbstractAction deleteAction, extractAction, biggerAction, smallerAction;

        ThumbPanel(boolean scrollfollow, boolean uselabels, int thumbnailsize, ThumbnailSidePanel parent) {
            super(null);
            this.scrollfollow = scrollfollow;
            this.uselabels = uselabels;
            this.thumbnailsize = thumbnailsize;
            this.parent = parent;
            this.pageNumberFormat = NumberFormat.getIntegerInstance();
            this.pagepanellist = Collections.synchronizedList(new ArrayList());  // Accessed from EDT and PainterThread - synchronize
            this.singlepagepanelfactory = new SinglePagePanelFactory(this, thumbnailsize);
            new DropTarget(this, this);
            setFocusable(true);

            this.deleteAction = new AbstractAction(UIManager.getString("PDFViewer.Delete")) {
                public void actionPerformed(ActionEvent event) {
                    if (JOptionPane.showConfirmDialog(docpanel, Util.getUIString("PDFViewer.ConfirmDeletePages", getSelectedPagesDescription()), getName(), JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
                        List l = new ArrayList();
                        for (int i=0;i<getComponentCount();i++) {
                            SinglePagePanel panel = (SinglePagePanel)getComponent(i);
                            if (panel.isSelected()) {
                                l.add(panel.getPage());
                                panel.setSelected(false);
                            }
                        }
                        final List oldpages = new ArrayList(pdf.getPages());
                        pdf.getPages().removeAll(l);
                        final List newpages = new ArrayList(pdf.getPages());
                        final PDFPage beforevisiblepage;
                        if (l.contains(docpanel.getPage())) {
                            docpanel.setPage(beforevisiblepage = pdf.getPage(0));
                        } else {
                            beforevisiblepage = docpanel.getViewport().getRenderingPage();
                        }
                        docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, new AbstractUndoableEdit() {
                            public String getPresentationName() {
                                return UIManager.getString("PDFViewer.Pages");
                            }
                            public boolean canUndo() {
                                return docpanel != null;
                            }
                            public boolean canRedo() {
                                return docpanel != null;
                            }
                            public void undo() {
                                super.undo();
                                pdf.getPages().clear();
                                pdf.getPages().addAll(oldpages);         // This makes list longer
                            }
                            public void redo() {
                                super.redo();
                                pdf.getPages().clear();
                                pdf.getPages().addAll(newpages);     // This makes list shorter and we
                                docpanel.setPage(beforevisiblepage); // may have deleted visible page
                            }
                        }));
                    }
                }
            };

            this.extractAction = new AbstractAction(UIManager.getString("PDFViewer.Extract")) {
                public void actionPerformed(ActionEvent event) {
                    if (JOptionPane.showConfirmDialog(docpanel, Util.getUIString("PDFViewer.ConfirmExtractPages", getSelectedPagesDescription()), getName(), JOptionPane.YES_NO_OPTION)==JOptionPane.YES_OPTION) {
                        JInternalFrame frame = (JInternalFrame)SwingUtilities.getAncestorOfClass(JInternalFrame.class, docpanel);
                        if (frame != null) {
                            List l = new ArrayList();
                            for (int i=0;i<getComponentCount();i++) {
                                SinglePagePanel panel = (SinglePagePanel)getComponent(i);
                                if (panel.isSelected()) {
                                    l.add(panel.getPage());
                                    panel.setSelected(false);
                                }
                            }
                            pdf.getPages().removeAll(l);
                            PDF newpdf = new PDF();
                            newpdf.getPages().addAll(l);
                            docpanel.setPage(pdf.getPage(0));
                            docpanel.getViewer().loadPDF(newpdf, "Pages from "+frame.getTitle());
                            docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, Undo.DISCARD));
                        }
                    }
                }
            };

            this.biggerAction = new AbstractAction(UIManager.getString("PDFViewer.Bigger")) {
                public void actionPerformed(ActionEvent event) {
                    singlepagepanelfactory.setSize((int)(singlepagepanelfactory.getSize() * 1.25));
                }
            };

            this.smallerAction = new AbstractAction(UIManager.getString("PDFViewer.Smaller")) {
                public void actionPerformed(ActionEvent event) {
                    singlepagepanelfactory.setSize((int)(singlepagepanelfactory.getSize() / 1.25));
                }
            };

            // Swing event model sucks - try doing this with InputMap
            addKeyListener(new KeyAdapter() {
                public void keyPressed(KeyEvent event) {
                    if (event.getKeyChar()=='+') {
                        if (biggerAction.isEnabled()) {
                            biggerAction.actionPerformed(new ActionEvent(ThumbPanel.this, ActionEvent.ACTION_PERFORMED, null));
                        }
                    } else if (event.getKeyChar()=='-') {
                        if (smallerAction.isEnabled()) {
                            smallerAction.actionPerformed(new ActionEvent(ThumbPanel.this, ActionEvent.ACTION_PERFORMED, null));
                        }
                    } else {
                        switch(event.getKeyCode()) {
                            case KeyEvent.VK_UP:
                            case KeyEvent.VK_DOWN:
                            case KeyEvent.VK_LEFT:
                            case KeyEvent.VK_RIGHT:
                                cursorKey(event);
                                break;
                            case KeyEvent.VK_DELETE:
                            case KeyEvent.VK_BACK_SPACE:
                                if (deleteAction.isEnabled()) {
                                    deleteAction.actionPerformed(new ActionEvent(ThumbPanel.this, ActionEvent.ACTION_PERFORMED, null));
                                }
                                break;
                        }
                    }
                }
            });
        }

        /**
         * Set the DocumentPanel this ThumbPanel relates to. This is really the
         * init method for this object - it starts the Animate and Painter thread
         * which continue running until this object is removed from a DocPanel,
         * even if there is no PDF.
         */
        void setDocumentPanel(DocumentPanel docpanel, boolean editable) {
            if (docpanel != this.docpanel) {
                if (this.docpanel != null) {
                    this.docpanel.removeDocumentPanelListener(this);
                }
                this.docpanel = docpanel;
                if (docpanel != null) {
                    docpanel.addDocumentPanelListener(this);
                    animatethread = new AnimateThread();
                    animatethread.start();
                    painterthread = new PainterThread();
                    painterthread.start();
                } else {
                    animatethread.interrupt();
                    painterthread.interrupt();
                }
            }
            initialize(docpanel == null ? null : docpanel.getPDF(), editable);
        }

        /**
         * Start or Stop the rendering of a PDF
         * 
         * Run on EDT
         */
        private void initialize(PDF pdf, boolean editable) {
            if (pdf == null && this.pdf != null) {
                this.pdf.removePropertyChangeListener(this);
                this.pdf = null;
                popup = null;
            } else if (pdf != null) {
                this.pdf = pdf;
                this.editable = editable && (pdf.getEncryptionHandler()==null || pdf.getEncryptionHandler().hasRight("Assemble"));
                pdf.addPropertyChangeListener(this);
                updatepagepanelsrequired = true;
                updatePagePanels(pdf);

                if (editable) {
                    popup = new JPopupMenu();
                    popup.add(deleteAction);
                    if (docpanel.getViewer() != null && docpanel.getViewer().hasFeature(MultiWindow.getInstance())) {
                        popup.add(extractAction);
                    }
                    popup.add(biggerAction);
                    popup.add(smallerAction);
                } else {
                    popup = null;
                }
            }
        }

        /**
         * The document has been updated - check for loading, closing
         * or changing the current page.
         */
        public void documentUpdated(DocumentPanelEvent event) {
            String type = event.getType();
            if (type == "loaded") {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        initialize(docpanel.getPDF(), editable);
                    }
                });
            } else if (type == "closing") {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        initialize(null, editable);
                    }
                });
            } else if (type == "pageChanged") {
                SinglePagePanel oldPagePanel = currentPagePanel;
                for (int i=0;i<pagepanellist.size();i++) {
                    currentPagePanel = (SinglePagePanel)pagepanellist.get(i);
                    if (currentPagePanel.getPage() == docpanel.getPage()) {
                        break;
                    }
                    currentPagePanel = null;
                }
                if (currentPagePanel != oldPagePanel) {
                    for (int i=0;i<getComponentCount();i++) {
                        SinglePagePanel panel = (SinglePagePanel)getComponent(i);
                        panel.setSelected(panel == currentPagePanel);
                    }
                    if (scrollfollow && currentPagePanel != null) {
                        scrollToDisplay(currentPagePanel.getBounds());
                    }
                }
            }
        }

        /**
         * The PDF fired a property change - document may have had its page
         * list altered.
         */
        public void propertyChange(PropertyChangeEvent event) {
            String name = event.getPropertyName();
            if (name.equals("pages") || name.equals("pagelabels")) {

                // Although we're most likely called on the EDT, schedule an
                // update for later anyway. This allows us to batch changes
                // that come in groups.
                updatepagepanelsrequired = true;
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        updatePagePanels(pdf);
                    }
                });
            }
        }

        /**
         * This method will add/remove the PagePanels so that there is a 1:1 mapping
         * from the page list to the pagepanel list. It should be called whenever
         * the PDF page list is updated.
         *
         * This is the only method that updates the "pagepanellist" array.
         *
         * Run on EDT
         *
         * @return true if the ordering of pages has changed
         */
        private void updatePagePanels(PDF pdf) {
            if (updatepagepanelsrequired) {
                updatepagepanelsrequired = false;
                synchronized(pagepanellist) {
                    this.pdf = pdf;
                    List newpages = pdf == null ? new ArrayList() : new ArrayList(pdf.getPages());
                    boolean changed = false, dimchanged = false;

                    for (int i=0;i<newpages.size();i++) {
                        PDFPage page = (PDFPage)newpages.get(i);
                        SinglePagePanel spp;
                        int oldix = -1;
                        for (int j=0;j<pagepanellist.size();j++) {
                            if (((SinglePagePanel)pagepanellist.get(j)).getPage() == page) {
                                oldix = j;
                                break;
                            }
                        }
                        if (oldix < 0) {        // New page
                            spp = singlepagepanelfactory.getSinglePagePanel(page);
                            if (i == getComponentCount()) {
                                add(spp);
                                pagepanellist.add(spp);
                            } else {
                                add(spp, i);
                                pagepanellist.add(i, spp);
                            }
                            dimchanged = true;
                            changed = true;
                        }
                    }

                    for (int i=0;i<pagepanellist.size();i++) {
                        SinglePagePanel spp = (SinglePagePanel)pagepanellist.get(i);
                        PDFPage page = spp.getPage();
                        int newix = newpages.indexOf(page);
                        if (newix < 0) {
                            spp = (SinglePagePanel)getComponent(i);
                            remove(spp);
                            pagepanellist.remove(i--);
                            changed = true;
                            dimchanged = true;
                            resetChildSize();
                        } else if (newix != i) {                // Order changed - swap
                            SinglePagePanel swap = (SinglePagePanel)pagepanellist.get(newix);
                            pagepanellist.remove(i);
                            pagepanellist.add(newix, spp);
                            add(spp, newix);
                            changed = true;
                            i--;
                        }
                    }
                    for (int i=0;i<pagepanellist.size();i++) {
                        SinglePagePanel spp = (SinglePagePanel)pagepanellist.get(i);
                        spp.resetPageNumber();
                    }

                    // Now order of components should match pages exactly
                    if (changed) {
                        if (dimchanged) {
                            childsize = null;
                        }
                        revalidate();
                        doLayout(getSize().width, false);
                        restartPainterThread();
                        repaint();
                        updateActions();
                    }
                }
            }
        }

        private String getSelectedPagesDescription() {
            int lastn = -2;
            StringBuffer sb = new StringBuffer();
            for (int i=0;i<=getComponentCount();i++) {
                if (i != getComponentCount() && ((SinglePagePanel)getComponent(i)).isSelected()) {
                    if (lastn < 0) {
                        lastn = i;
                    }
                } else {
                    if (lastn >= 0) {
                        if (sb.length()>0) {
                            sb.append(",");
                        }
                        sb.append(lastn+1);
                        if (i != lastn + 1) {
                            sb.append("-");
                            sb.append(i);
                        }
                    }
                    lastn = -2;
                }
            }
            return sb.toString();
        }


        /**
         * Signal the PainterThread that >=1 of the pages has changed and
         * may need repainting
         */
        void restartPainterThread() {
            painterthread.restart();
        }

        /**
         * Called when >=1 of the pages has changed its dimensions, which
         * requires laying out the children again
         */
        void resetChildSize() {
            childsize = null;
            invalidate();
            revalidate();
            repaint();
        }

        public void doLayout() {
            // Called often, even when pane is hidden, so be quick.
            int w = getSize().width;
            if (w > 0) {
                doLayout(getSize().width, false);
            }
        }

        /**
         * Calculate the dimensions of this ThumbPanel and optionally move or
         * animate the panels to their correct positions.
         *
         * Run on EDT
         * 
         * @param w the width that the ThumbPanel is (will be)
         * @param sizingonly if true, don't move anything - we're just measuring
         * @return the (anticipated) size of the ThumbPanel given the specified width
         */
        private Dimension doLayout(int w, boolean sizingonly) {
            Dimension cs = getChildSize();
            int cx = 0, cy = 0;
            int xmax = 0, ymax = 0;
            int n = getComponentCount();
            Map animate = null;
            for (int i = 0; i < n; i++) {
                SinglePagePanel c = (SinglePagePanel)getComponent(i);
                if (!sizingonly) {
                    Rectangle src = animatethread.getDestinationBounds(c);
                    Rectangle dst = new Rectangle(cx, cy, cs.width, cs.height);
                    if (!animatemovements || src.width == 0) {
                        c.setBounds(dst);
                        revalidate();
                        repaint();
                    } else if (!src.equals(dst)) {
                        if (animate == null) {
                            animate = new HashMap();
                        }
                        animate.put(c, new Rectangle[] { src, dst });
                    }
                }
                xmax = Math.max(xmax, cx + cs.width);
                ymax = Math.max(ymax, cy + cs.height);
                cx += cs.width;
                if (cx + cs.width > w) {
                    cx = 0;
                    cy += cs.height;
                }
            }
            if (animate != null) {
                animatethread.go(animate);
            }
            if (!sizingonly) {
                animatemovements = true;
            }

            return new Dimension(xmax, ymax);
        }

        /**
         * Thread that animates components moving from one position to another.
         */
        private class AnimateThread extends Thread implements SoftInterruptibleThread {

            private Map map;
            private volatile boolean interrupt, waiting;

            AnimateThread() {
                super("ThumbnailAnimator");
                setDaemon(true);
            }

            public void interrupt() {
                interrupt = true;
                if (waiting) {
                    // Don't want to do this unless we have to, interrupt()
                    // closes the stream, so no further access to the PDF
                    // object is possible if we're reading it with NIO.
                    super.interrupt();
                }
            }

            public boolean isSoftInterrupted() {
                return interrupt;
            }

            /**
             * Return the bounds of the specific component when the current
             * animation (if any) is finished. In the default case this will
             * just be the bounds of the Component.
             */
            synchronized Rectangle getDestinationBounds(Component c) {
                Rectangle[] r = map == null ? null : (Rectangle[])map.get(c);
                return r == null ? c.getBounds() : r[1];
            }

            /**
             * Start a new animation
             * @param map a Map of Component -> [FromRectangle, ToRectangle]
             */
            public synchronized void go(Map map) {
                this.map = map;
                notifyAll();
            }

            public void run() {
                while (docpanel != null) {
                    Map map;
                    synchronized(this) {
                        if (pdf == null || (map = this.map) == null) {
                            try {
                                waiting = true;
                                wait();
                            } catch (InterruptedException e) {}
                            waiting = false;
                            continue;
                        }
                    }
                    int steps = 10;
                    for (int i=0;i<=steps;i++) {
                        for (Iterator j = map.entrySet().iterator();j.hasNext();) {
                            Map.Entry e = (Map.Entry)j.next();
                            SinglePagePanel panel = (SinglePagePanel)e.getKey();
                            if (panel.getParent() != null) {
                                Rectangle[] r = (Rectangle[])e.getValue();
                                int x = r[0].x + (r[1].x - r[0].x)*i/steps;
                                int y = r[0].y + (r[1].y - r[0].y)*i/steps;
                                int w = r[0].width + (r[1].width - r[0].width)*i/steps;
                                int h = r[0].height + (r[1].height - r[0].height)*i/steps;
                                panel.setBounds(x, y, w, h);
                                repaint();
                            } else {
                                j.remove();
                            }
                        }
                        try {
                            Thread.sleep(15);
                        } catch (InterruptedException e) { }
                    }
                    synchronized(this) {
                        if (map == this.map) {
                            this.map = null;
                        }
                    }
                }
            }

        }

        /**
         * Thread that cycles through all the SinglePagePanels and paints
         * them if necessary.
         */
        private class PainterThread extends Thread implements SoftInterruptibleThread {

            private volatile int centerpage;
            private volatile boolean paused, completed, interrupt, waiting;

            PainterThread() {
                super("ThumbnailPainter");
                setDaemon(true);
                setPriority(Thread.MIN_PRIORITY+2);
            }

            public void interrupt() {
                interrupt = true;
                if (waiting) {
                    // Don't want to do this unless we have to, interrupt()
                    // closes the stream, so no further access to the PDF
                    // object is possible if we're reading it with NIO.
                    super.interrupt();
                }
            }

            public boolean isSoftInterrupted() {
                return interrupt;
            }


            boolean isComplete() {
                return this.completed;
            }

            synchronized void restart() {
                this.completed = false;
                notifyAll();
            }

            synchronized void setPaused(boolean paused) {
                this.paused = paused;
                notifyAll();
            }

            boolean isPaused() {
                return paused;
            }

            /**
             * Set which component is in the center of the viewport
             */
            void setCenterComponent(int page) {
                this.centerpage = page;
            }

            public void run() {
                while (docpanel != null) {
                    synchronized(this) {
                        if (pdf == null || paused || completed) {
                            try {
                                waiting = true;
                                wait();
                            } catch (InterruptedException e) {}
                            waiting = false;
                            continue;
                        }
                    }

                    // Start at centerpage, then move up, down, up down based on d=delta
                    // d goes -1 2 -3 4... or 1 -2 3 -4... to cover [v v-1 v+1 v-2 v+2 ...]
                    int v = centerpage;
                    int d = 1;

                    SinglePagePanel spp = null;
                    while (v >= 0 && v < pagepanellist.size() && !(spp=(SinglePagePanel)pagepanellist.get(v)).isRendered() && !spp.isRendering()) {
                        v += d;
                        d = d < 0 ? -d + 1 : -d - 1;
                        if (v < 0 || v >= pagepanellist.size()) {
                            v += d;
                            d = d < 0 ? -d + 1 : -d - 1;
                        }
                        spp = null;
                    }
                    if (spp != null) {
                        spp.drawImage();
                    } else {
                        completed = true;
                    }
                }
            }

        }

        /**
         * Return true if the PDF in the panel can be edited,
         * false if this ThumbPanel is read-only
         */
        boolean isEditable() {
            return popup != null;
        }

        boolean isRunning() {
            return painterthread != null && !painterthread.isComplete();
        }

        /**
         * Get the DocumentPanel this ThumbPanel relates to
         */
        public DocumentPanel getDocumentPanel() {
            return docpanel;
        }

        boolean isPaused() {
            return painterthread.isPaused();
        }

        void setPaused(boolean pause) {
            if (pause != isPaused()) {
                painterthread.setPaused(pause);
                if (pause) {
                    synchronized(pagepanellist) {
                        for (int i=0;i<pagepanellist.size();i++) {
                            SinglePagePanel panel = (SinglePagePanel)pagepanellist.get(i);
                            synchronized(panel) {
                                if (panel.isRendering()) {
                                    try {
                                        panel.wait();
                                    } catch (InterruptedException e) {}
                                }
                            }
                        }
                    }
                }
            }
        }

        private void cursorKey(KeyEvent event) {
            Dimension cs = getChildSize();
            Dimension s = getSize();
            int rowlen = s.width / cs.width;
            int num = -1;
            for (int i=0;i<getComponentCount();i++) {
                if (getComponent(i) == currentPagePanel) {
                    num = i;
                    break;
                }
            }
            switch (event.getKeyCode()) {
            case KeyEvent.VK_UP:
                num -= rowlen;
                break;
            case KeyEvent.VK_DOWN:
                num += rowlen;
                break;
            case KeyEvent.VK_RIGHT:
                num++;
                break;
            case KeyEvent.VK_LEFT:
                num--;
                break;
            }
            if (num >= 0 && num < getComponentCount()) {
                SinglePagePanel spp = (SinglePagePanel)getComponent(num);
                spp.selectAction(event);
            }
        }

        private void scrollToDisplay(Rectangle kid) {
            // Don't be so eager to scroll - looks better on Windows
            Rectangle vis = getVisibleRect();
            if (kid.y < vis.y || kid.y + kid.height > vis.y + vis.height) {
                scrollRectToVisible(kid);
            }
        }

        //----------------------------------------------------------------------------
        // Drag/Drop methods
        //----------------------------------------------------------------------------

        public void dragOver(DropTargetDragEvent event) {
            dragEnter(event);
        }

        /**
         * Accept or Reject the event based on its DataFlavors
         */
        public void dragEnter(DropTargetDragEvent event) {
            if (isEditable() && (event.getDropAction() & DnDConstants.ACTION_COPY_OR_MOVE) != 0) {
                DataFlavor[] flavors = event.getCurrentDataFlavors();
                for (int i=0;i<flavors.length;i++) {
                    if (flavors[i].isMimeTypeEqual(DataFlavor.javaJVMLocalObjectMimeType) || DragAndDrop.canImport(flavors[i])) {
                        dragTarget = findComponentAt(event.getLocation());
                        event.acceptDrag(DnDConstants.ACTION_COPY_OR_MOVE);
                        repaint();
                        return;
                    }
                }
            }
            dragTarget = null;
            event.rejectDrag();
            repaint();
        }

        public void dragExit(DropTargetEvent event) {
            dragTarget = null;
            repaint();
        }

        /**
         * Accept a Drop. This method will update the PDF page list, which will fire
         * an event which this class is listning for.
         */
        public void drop(DropTargetDropEvent event) {
            if (isEditable() && dragTarget != null) {
                int index = dragTarget == this ? pagepanellist.size() - 1 : pagepanellist.indexOf(dragTarget);
                try {
                    final Transferable transfer = event.getTransferable();
                    if (transfer.isDataFlavorSupported(SinglePagePanel.DRAGGABLE_PAGE)) {
                        event.acceptDrop(DnDConstants.ACTION_MOVE);
                        SinglePagePanel source = (SinglePagePanel)transfer.getTransferData(SinglePagePanel.DRAGGABLE_PAGE);
                        if (source != pagepanellist.get(index)) {
                            final ThumbPanel otherthumbpanel = source.thumbpanel;
                            List sourcepages = new ArrayList();
                            for (int i=0;i<otherthumbpanel.getComponentCount();i++) {
                                SinglePagePanel spp = (SinglePagePanel)otherthumbpanel.getComponent(i);
                                if (spp.isSelected()) {
                                    sourcepages.add(spp.getPage());
                                    spp.setSelected(spp == source);
                                }
                            }

                            List otherpages = otherthumbpanel.getDocumentPanel().getPDF().getPages();
                            if (otherthumbpanel != this) {
                                // We're moving pages between active DocumentPanels!
                                if (sourcepages.size() == otherthumbpanel.getComponentCount()) {
                                    otherthumbpanel.getDocumentPanel().setDirty(false);
                                    PDFViewer viewer = otherthumbpanel.getDocumentPanel().getViewer();
                                    viewer.closeDocumentPanel(otherthumbpanel.getDocumentPanel());
                                    otherpages = null;
                                } else if (sourcepages.contains(otherthumbpanel.docpanel.getPage())) {
                                    // Got to ensure other DocumentPanel isn't displaying a page we're
                                    // removing, or things go pearshaped.
                                    for (int i=0;i<otherpages.size();i++) {
                                        PDFPage otherpage = (PDFPage)otherpages.get(i);
                                        if (!sourcepages.contains(otherpage)) {
                                            otherthumbpanel.docpanel.setPage(otherpage);
                                            break;
                                        }
                                    }
                                }
                            }

                            List pages = pdf.getPages();
                            final List oldpages = new ArrayList(pages);
                            if (otherpages != null) {
                                otherpages.removeAll(sourcepages);
                            }
                            pages.addAll(Math.min(pages.size(), index), sourcepages);
                            final List newpages = new ArrayList(pages);

                            if (otherthumbpanel == this) {
                                // We can't really undo when moving pages between docs, because
                                // if the other document is closed after we've dragged its pages
                                // in here, this even will still hold a reference to it via all
                                // it's page objects, which could be expensive.
                                //
                                docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, new AbstractUndoableEdit() {
                                    final PDFPage beforevisiblepage = docpanel.getPage();
                                    public String getPresentationName() {
                                        return UIManager.getString("PDFViewer.Pages");
                                    }
                                    public boolean canUndo() {
                                        return docpanel != null;
                                    }
                                    public boolean canRedo() {
                                        return docpanel != null;
                                    }
                                    public void undo() {
                                        super.undo();
                                        pdf.getPages().clear();
                                        pdf.getPages().addAll(oldpages);         // This makes list shorter, and we
                                        docpanel.setPage(beforevisiblepage); // may have deleted visible page
                                    }
                                    public void redo() {
                                        super.redo();
                                        pdf.getPages().clear();
                                        pdf.getPages().addAll(newpages);        // This makes list longer
                                    }
                                }));
                            } else {
                                docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, Undo.DISCARD));
                            }
                        }
                        event.dropComplete(true);
                    } else if (DragAndDrop.canImport(transfer.getTransferDataFlavors())) {
                        event.acceptDrop(DnDConstants.ACTION_COPY);
                        // Create a list of files by using a DragAndDrop to process the Transferable.
                        // Then create a task to load and process that list, adding the pages to
                        // this document.
                        final List list = new ArrayList();
                        DragAndDrop dnd = new DragAndDrop() {
                            public void action(File file) {
                                list.add(file);
                            }
                            public void action(java.net.URL url) { }
                        };
                        final PDFViewer viewer = docpanel.getViewer();
                        if (dnd.processTransferable(transfer) && list.size() > 0 && viewer != null) {
                            final ViewerFeature[] features = viewer.getFeatures();
                            final int[] targetindex = new int[] { index };
                            final List oldpages = new ArrayList(pdf.getPages());

                            LongRunningTask task = new LongRunningTask() {
                                private Importer.ImporterTask task;
                                public void cancel() {
                                    super.cancel();
                                    if (task != null) {
                                        task.cancel();
                                    }
                                }
                                public boolean isCancellable() {
                                    return true;
                                }
                                public float getProgress() {
                                    return task==null ? 0 : task.getProgress();
                                }
                                private Importer.ImporterTask getImporterTask(File file) throws IOException {
                                    for (int j=0;j<features.length;j++) {
                                        if (features[j] instanceof Importer && ((Importer)features[j]).matches(file)) {
                                            Importer importer = (Importer)features[j];
                                            if (importer instanceof PDFImporter) {
                                                ((PDFImporter)importer).setUseInputStream(true);
                                            }
                                            return importer.getImporter(viewer, file);
                                        }
                                    }
                                    return null;
                                }
                                public void run() throws IOException {
                                    for (int i=0;i<list.size();i++) {
                                        File file = (File)list.get(i);
                                        task = getImporterTask(file);
                                        if (task != null) {
                                            final PDF newpdf = task.loadPDF();
                                            if (!isCancelled() && newpdf != null) {
                                                final List newpagelist = new ArrayList(newpdf.getPages());
                                                final int offset = targetindex[0];
                                                targetindex[0] += newpagelist.size();

                                                SwingUtilities.invokeLater(new Runnable() {
                                                    public void run() {
                                                        pdf.getPages().addAll(offset, newpagelist);
                                                    }
                                                });
                                            }
                                        }
                                    }

                                    SwingUtilities.invokeLater(new Runnable() {
                                        public void run() {
                                            docpanel.fireUndoableEditEvent(new UndoableEditEvent(docpanel, new AbstractUndoableEdit() {
                                                final List newpages = new ArrayList(pdf.getPages());
                                                final PDFPage beforevisiblepage = docpanel.getPage();
                                                public String getPresentationName() {
                                                    return UIManager.getString("PDFViewer.Pages");
                                                }
                                                public boolean canUndo() {
                                                    return docpanel != null;
                                                }
                                                public boolean canRedo() {
                                                    return docpanel != null;
                                                }
                                                public void undo() {
                                                    super.undo();
                                                    pdf.getPages().clear();
                                                    pdf.getPages().addAll(oldpages);     // This makes list shorter, and we
                                                    docpanel.setPage(beforevisiblepage); // may have deleted visible page
                                                }
                                                public void redo() {
                                                    super.redo();
                                                    pdf.getPages().clear();
                                                    pdf.getPages().addAll(newpages);        // This makes list longer
                                                }
                                            }));
                                        }
                                    });
                                }
                            };
                            task.start(viewer, UIManager.getString("PDFViewer.Loading"));
                        }
                        event.dropComplete(true);
                    } else {
                        event.rejectDrop();
                    }
                } catch (Throwable  e) {
                    Util.displayThrowable(e, docpanel);
                    event.rejectDrop();
                }
            } else {
                event.rejectDrop();
            }
            dragTarget = null;
            repaint();
        }

        public void dropActionChanged(DropTargetDragEvent event) {
        }

        //----------------------------------------------------------------------------
        // Popup methods
        //----------------------------------------------------------------------------

        JPopupMenu getPopup() {
            return popup;
        }

        /** 
         * Update menu depending on how many items are selected - can't delete all pages!
         */
        void updateActions() {
            int select = 0, total = 0;
            for (int i=0;i<getComponentCount();i++) {
                SinglePagePanel panel = (SinglePagePanel)getComponent(i);
                if (panel.isSelected()) {
                    select++;
                }
                total++;
            }
            deleteAction.setEnabled(popup != null && select > 0 && select < total);
            extractAction.setEnabled(popup != null && select > 0 && select < total);
        }

        //----------------------------------------------------------------------------
        // Scrollable Methods
        //----------------------------------------------------------------------------

        public void setBounds(int x, int y, int width, int height) {
            Component parent = getParent();
            if (parent != null && painterthread != null) {
                Dimension p = parent.getSize();

                // To ensure we fill our scrollpane parent entirely
                height = Math.max(height, p.height);
                width = Math.max(width, p.width);

                // height = height of overall panel
                // p.height = height of visible portion
                // y = 0 when scrolled to top, (height - p.height) when scrolled to bottom
                int cy = (p.height/2) - y; // y at center of viewport (0..height)
                painterthread.setCenterComponent(cy * getComponentCount() / height);
            }
            Dimension size = getSize();
            if (width != size.width) {
                animatemovements = false;           // Don't animate on resize
            }
            super.setBounds(x, y, width, height);
        }

        /**
         * Returns the maximum width and height of all children.
         */
        private synchronized Dimension getChildSize() {
            if (childsize == null) {
                childsize = new Dimension(0, 0);
                int n = getComponentCount();
                // Determine max child width and height
                for (int i = 0; i < n; i++) {
                    Component c = getComponent(i);
                    Dimension cs = c.getPreferredSize();
                    if (cs.width > childsize.width) {
                        childsize.width = cs.width;
                    }
                    if (cs.height > childsize.height) {
                        childsize.height = cs.height;
                    }
                }
            }
            return childsize;
        }

        public Dimension getPreferredSize() {
            Container parent = getParent();
            if (parent == null || parent.getBounds().width <= 0) {
                Dimension cs = getChildSize();
                return new Dimension(cs.width, cs.height * getComponentCount());
            } else {
                return doLayout(parent.getBounds().width, true);
            }
        }

        public Dimension getPreferredScrollableViewportSize() {
            return getPreferredSize();
        }

        public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {
            int val = getScrollableBlockIncrement(visibleRect, orientation, direction) / 100;
            return val==0 ? 1 : val;
        }

        public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction) {
            return orientation == SwingConstants.VERTICAL ? getParent().getHeight() : getParent().getWidth();
        }

        public boolean getScrollableTracksViewportWidth() {
            return true;
        }

        public boolean getScrollableTracksViewportHeight() {
            return false;
        }

        public void paintComponent(Graphics g) {
            super.paintComponent(g);
            if (dragTarget != null) {
                boolean end = dragTarget == this;
                Component c = end ? getComponent(getComponentCount() - 1) : dragTarget;

                if (c != null) {
                    // Determine whether all components are arranged vertically
                    boolean vertical = true;
                    for (int i=0;vertical && i < getComponentCount();i++) {
                        vertical &= getComponent(i).getLocation().x == 0;
                    }

                    Color selection = UIManager.getColor("List.selectionBackground");
                    if (selection == null) {
                        selection = new Color(153, 153, 204);
                    }
                    g.setColor(selection);

                    Rectangle cb = c.getBounds();
                    if (vertical) {
                        // Draw a horizontal insertion point
                        if (end) {
                            g.fillRect(cb.x, cb.y + cb.height - 1, cb.width, 2);
                        } else {
                            g.fillRect(cb.x, cb.y, cb.width, 2);
                        }
                    } else {
                        // Draw a vertical insertion point
                        if (end) {
                            g.fillRect(cb.x + cb.width - 2, cb.y, 2, cb.height);
                        } else {
                            g.fillRect(cb.x, cb.y, 2, cb.height);
                        }
                    }
                }
            }
        }

        //----------------------------------------------------------------------------
        // Mouse Methods
        //----------------------------------------------------------------------------

        public void autoscroll(Point p) {
            JScrollPane scrollpane = (JScrollPane)SwingUtilities.getAncestorOfClass(JScrollPane.class, this);
            if (scrollpane != null) {
                JScrollBar bar = scrollpane.getVerticalScrollBar();
                Rectangle r = getVisibleRect();
                if (p.y <= r.y + 50) {
                    int diff = r.y + 50 - p.y;
                    bar.setValue(bar.getValue() - diff*5);
                }
                if (p.y >= r.y + r.height - 50) {
                    int diff = p.y - (r.y + r.height - 50);
                    bar.setValue(bar.getValue() + diff*5);
                }
            }
        }

        public Insets getAutoscrollInsets() {
            Rectangle r = getVisibleRect();
            Dimension size = getSize();
            int border = 50;
            return new Insets(r.y+border, r.x, size.height-r.y-r.height+border, size.width-r.x-r.width);
        }
    }

    /**
     * Factory to create SinglePaegPanels which will cache, to allow
     * pages to be removed and readded without being discarded. Necessary
     * as moving a page in a PDF often involves taking it temporarily
     * out of the page tree.
     */
    static class SinglePagePanelFactory {

        private final ThumbPanel thumbpanel;
        private final WeakHashMap map;
        private int thumbnailsize;

        SinglePagePanelFactory(ThumbPanel thumbpanel, int thumbnailsize) {
            this.thumbpanel = thumbpanel;
            this.thumbnailsize = thumbnailsize;
            this.map = new WeakHashMap();
        }

        synchronized SinglePagePanel getSinglePagePanel(PDFPage page) {
            SinglePagePanel panel = (SinglePagePanel)map.get(page);
            if (panel == null) {
                map.put(page, panel = new SinglePagePanel(thumbpanel, page));
                panel.setThumbnailSize(thumbnailsize);
            }
            return panel;
        }

        int getSize() {
            return thumbnailsize;
        }

        void setSize(int size) {
            if (size != thumbnailsize && size >= 20 && size <= 200) {
                this.thumbnailsize = size;
                for (Iterator i = map.values().iterator();i.hasNext();) {
                    SinglePagePanel panel = (SinglePagePanel)i.next();
                    if (panel.getParent() != null) {
                        panel.setThumbnailSize(thumbnailsize);
                    } else {
                        i.remove();
                    }
                }
            }
        }

    }

    /**
     * JPanel representing a single page image.
     */
    static class SinglePagePanel extends JPanel implements MouseListener, DragSourceListener, DragGestureListener, Transferable, PropertyChangeListener {

        private static final int MARGIN = 3;
        private static final int SHADOWDEPTH = 1;
        private static final Font PAGENUMBERFONT = new Font("SansSerif", Font.PLAIN, 9);
        static final DataFlavor DRAGGABLE_PAGE = new DataFlavor(DataFlavor.javaJVMLocalObjectMimeType, "SinglePagePanel");

        private final ThumbPanel thumbpanel;
        private final Dimension imagesize;
        private int thumbnailsize;
        private float dpi;
        private PagePainter painter;
        private String pagenumberstring;
        private boolean selected;
        private volatile BufferedImage image, oldimage;
        private volatile boolean isrendering;

        SinglePagePanel(ThumbPanel thumbpanel, PDFPage page) {
            this.thumbpanel = thumbpanel;
            this.painter = thumbpanel.docpanel.getParser().getPagePainter(page);
            addMouseListener(this);
            DragSource ds = DragSource.getDefaultDragSource();
            ds.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_MOVE, this);
            setOpaque(false);
            imagesize = new Dimension();

            if (thumbpanel.isEditable()) {
                setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
            } else {
                setCursor(null);
            }
        }

        /**
         * Resize the SinglePagePanel
         */
        void setThumbnailSize(int thumbnailsize) {
            this.thumbnailsize = thumbnailsize;
            painter.interrupt();
            painter = thumbpanel.docpanel.getParser().getPagePainter(getPage());     // Shouldn't have to do this, but otherwise dimensions are all wrong on rotate.
            isrendering = false;
            PDFPage page = getPage();
            float[] box = page.getBox("ViewBox");
            float mul = thumbnailsize / Math.max(box[2]-box[0], box[3]-box[1]);
            imagesize.width = (int)Math.floor((box[2]-box[0]) * mul);
            imagesize.height = (int)Math.floor((box[3]-box[1]) * mul);
            dpi = Math.min(imagesize.width / (box[2]-box[0]), imagesize.height / (box[3]-box[1])) * 72;
            int pad = 2 + SHADOWDEPTH + MARGIN + MARGIN;
            image = null;
            setPreferredSize(new Dimension(imagesize.width + pad, imagesize.height + pad + PAGENUMBERFONT.getSize() + 1));
            thumbpanel.resetChildSize();
            thumbpanel.restartPainterThread();
        }

        PDFPage getPage() {
            return painter.getPage();
        }

        void resetPageNumber() {
            pagenumberstring = null;
            repaint();
        }

        public void addNotify() {
            super.addNotify();
            PDFPage page = getPage();
            page.addPropertyChangeListener(this);
            List annots = page.getAnnotations();
            for (int i=0;i<annots.size();i++) {
                PDFAnnotation a = (PDFAnnotation)annots.get(i);
                a.addPropertyChangeListener(this);
            }
        }

        public void removeNotify() {
            super.removeNotify();
            painter.interrupt();
            PDFPage page = getPage();
            page.removePropertyChangeListener(this);
            List annots = page.getAnnotations();
            for (int i=0;i<annots.size();i++) {
                PDFAnnotation a = (PDFAnnotation)annots.get(i);
                a.removePropertyChangeListener(this);
            }
            isrendering = false;
        }

        public void propertyChange(PropertyChangeEvent event) {
            String name = event.getPropertyName();
            if (name.endsWith("box") || name.equals("orientation")) {
                setThumbnailSize(thumbnailsize);
            } else {
                if (name.equals("annotations") && event.getSource() == getPage()) {
                    Object oldval = event.getOldValue();
                    Object newval = event.getNewValue();
                    if (oldval instanceof PDFAnnotation) {
                        ((PDFAnnotation)oldval).removePropertyChangeListener(this);
                    }
                    if (newval instanceof PDFAnnotation) {
                        ((PDFAnnotation)newval).addPropertyChangeListener(this);
                    }
                }
                oldimage = image;
                image = null;
                painter.interrupt();
                isrendering = false;
                thumbpanel.restartPainterThread();
            }
        }

        /**
         * Mark this page as selected
         */
        void setSelected(boolean selected) {
            if (selected != this.selected) {
                this.selected = selected;
                repaint();
            }
        }

        /**
         * Return true if this page is selected
         */
        boolean isSelected() {
            return selected;
        }

        /**
         * Return true if the page image is currently rendering
         */
        boolean isRendering() {
            return isrendering;
        }

        /**
         * Return true if this page needs to be redrawn
         */
        boolean isRendered() {
            return image == null;
        }

        /**
         * Draw the image if it's not already drawn.
         */
        synchronized void drawImage() {
            if (image == null) {
                isrendering = true;
                image = painter.getImage(dpi);
                if (image != null) {
                    oldimage = null;
                }
                isrendering = false;
                notifyAll();
                repaint();
            }
        }

        public void paintComponent(Graphics g) {
            super.paintComponent(g);
            Color foreground = UIManager.getColor("List.foreground");
            if (foreground == null) {
                foreground = Color.black;
            }
            Color background = UIManager.getColor("List.background");
            if (foreground == null) {
                foreground = new Color(204, 204, 204);
            }

            Dimension size = getSize();
            if (isSelected()) {
                background = UIManager.getColor("List.selectionBackground");
                if (background == null) {
                    background = new Color(153, 153, 204);
                }
                g.setColor(background);
                g.fillRect(0, 0, getWidth(), getHeight());
            }

            // Calculate image border decoration bounds
            int iw = imagesize.width + 2;
            int ih = imagesize.height + 2;
            int ix = (size.width - iw) / 2;
            int iy = (((size.height - PAGENUMBERFONT.getSize()) - 1) - ih) / 2;

            // Draw shadow
            Color shadow = getIntermediateColor(background, foreground);
            g.setColor(shadow);
            g.fillRect(ix + SHADOWDEPTH, iy + SHADOWDEPTH, iw, ih);

            // Draw image
            BufferedImage image = this.image;
            if (image == null) {
                image = this.oldimage;
            }
            if (image != null) {
                ((Graphics2D)g).drawRenderedImage(image, AffineTransform.getTranslateInstance(ix+1, iy+1));
            } else {
                g.setColor(Color.gray);
                g.fillRect(ix+1, iy+1, imagesize.width, imagesize.height);
            }

            // Draw solid border
            g.setColor(foreground);
            g.drawRect(ix, iy, iw, ih);

            if (pagenumberstring == null) {
                int pagenumber = getPage().getPageNumber();
                pagenumberstring = thumbpanel.uselabels ? getPage().getPDF().getPageLabel(pagenumber - 1) : null;
                if (pagenumberstring == null) {
                    pagenumberstring = thumbpanel.pageNumberFormat.format(pagenumber);
                }
            }
            // Draw page number
            g.setColor(foreground);
            g.setFont(PAGENUMBERFONT);
            FontRenderContext frc = ((Graphics2D)g).getFontRenderContext();
            Rectangle2D nw2d = PAGENUMBERFONT.getStringBounds(pagenumberstring, frc);
            int nw = (int) nw2d.getWidth();
            int nx = (size.width - nw) / 2;
            int ny = size.height - MARGIN;
            g.drawString(pagenumberstring, nx, ny);
        }

        static Color getIntermediateColor(Color c1, Color c2) {
            int r1 = c1.getRed(), g1 = c1.getGreen(), b1 = c1.getBlue();
            int r2 = c2.getRed(), g2 = c2.getGreen(), b2 = c2.getBlue();
            int r3 = (r1 < r2) ? r1 + ((r2 - r1) / 2) : r2 + ((r1 - r2) / 2);
            int g3 = (g1 < g2) ? g1 + ((g2 - g1) / 2) : g2 + ((g1 - g2) / 2);
            int b3 = (b1 < b2) ? b1 + ((b2 - b1) / 2) : b2 + ((b1 - b2) / 2);
            return new Color(r3, g3, b3);
        }

        public void mouseEntered(MouseEvent event) { }
        public void mouseExited(MouseEvent event) { }
        public void mouseClicked(MouseEvent event) { }
        public void mouseReleased(MouseEvent event) {
            if (event.isPopupTrigger() && thumbpanel.getPopup() != null) {
                thumbpanel.getPopup().show(this, event.getX(), event.getY());
            } else {
                selectAction(event);
            }
        }

        public void mousePressed(MouseEvent event) {
            if (event.isPopupTrigger() && thumbpanel.getPopup() != null) {
                thumbpanel.getPopup().show(this, event.getX(), event.getY());
            } else {
                selectAction(event);
            }
        }

        void selectAction(InputEvent event) {
            int mod = event.getModifiers();
            boolean shift = (mod&event.SHIFT_MASK)!=0;
            boolean ctrl = (mod&(event.CTRL_MASK+event.META_MASK))!=0;

            if (!shift && !ctrl) {
                // Do this so we can update menu, then set page.
                for (int i=0;i<thumbpanel.getComponentCount();i++) {
                    SinglePagePanel spp = (SinglePagePanel)thumbpanel.getComponent(i);
                    spp.setSelected(spp == this);
                }
                thumbpanel.docpanel.setPage(getPage());
                thumbpanel.currentPagePanel = this;
            } else if (ctrl) {
                setSelected(!isSelected());
                thumbpanel.currentPagePanel = this;
            } else if (shift) {
                DocumentPanel docpanel = thumbpanel.docpanel;
                PDFPage currentpage = thumbpanel.docpanel.getPage();
                if (currentpage == null) {
                    currentpage = docpanel.getViewport().getRenderingPage();
                }
                if (currentpage != null) {
                    boolean toggle = false;
                    for (int i=0;i<thumbpanel.getComponentCount();i++) {
                        SinglePagePanel spp = (SinglePagePanel)thumbpanel.getComponent(i);
                        spp.setSelected(toggle);
                        if (spp.getPage() == currentpage) {
                            spp.setSelected(true);
                            toggle = !toggle;
                        }
                        if (spp == this) {
                            spp.setSelected(true);
                            toggle = !toggle;
                        }
                    }
                }
                thumbpanel.currentPagePanel = this;
            }
            thumbpanel.scrollToDisplay(getBounds());
            thumbpanel.updateActions();
        }

        public void dragGestureRecognized(DragGestureEvent event) {
            if (thumbpanel.isEditable()) {
                if (!isSelected()) {
                    selectAction((MouseEvent)event.getTriggerEvent());
                }
                try {
                    event.startDrag(null, this, this);
                } catch (InvalidDnDOperationException e) {
                    // drag already in progress, ignore - nothing we can do at this point, drag/drop is jammed
                }
            }
        }

        public Object getTransferData(DataFlavor flavor) {
            return flavor == DRAGGABLE_PAGE ? this : null;
        }

        public DataFlavor[] getTransferDataFlavors() {
            return new DataFlavor[] { DRAGGABLE_PAGE };
        }

        public boolean isDataFlavorSupported(DataFlavor flavor) {
            return flavor == DRAGGABLE_PAGE;
        }

        public void dragDropEnd(DragSourceDropEvent event) { }
        public void dragEnter(DragSourceDragEvent event) { }
        public void dragExit(DragSourceEvent event) { }
        public void dragOver(DragSourceDragEvent event) { }
        public void dropActionChanged(DragSourceDragEvent event) { }

        public String toString() {
            return super.toString() + " P"+(getPage().getPageNumber()-1);
        }


    }

}
